import type {
  UserListResponseModel,
  ListViewConfig,
  AddViewConfig,
  UserDetailResponseModel,
  LabeledValue,
  FormValues,
} from '@/helpers/models';
import {fetchAuthenticated} from '@/helpers/retrieveData';
import {useErrorStore} from '@/stores/error';
import i18next from 'i18next';

const fetchListViewConfigMocked = (): ListViewConfig => {
  const _config: ListViewConfig = {
    url: 'str',
    allowedGlobalActions: ['add'],
    searchFormOptions: {
      schools: [
        {
          value: 'DEMOSCHOOL',
          label: 'DEMOSCHOOL',
        }, {
          value: 'School1',
          label: 'Schule 1',
        },
      ],
      activationStates: [
        {
          value: 'aktiviert',
          label: 'Aktiviert',
        }, {
          value: 'deactivated',
          label: 'Deaktiviert',
        },
      ],
      networkAccess: [
        {
          value: 'aktiviert',
          label: 'Aktiviert',
        }, {
          value: 'deactivated',
          label: 'Deaktiviert',
        },
      ],
      roles: [
        {
          value: 'student',
          label: 'Lernende',
        },
        {
          value: 'generic_user',
          label: 'Nutzende',
        },
        {
          value: 'teacher',
          label: 'Lehrende',
        },
        {
          value: 'school_admin',
          label: 'Schuladmins',
        },
        {
          value: 'technical_admin',
          label: 'Fachliche Leitstelle',
        },
      ],
    },
    columns: [
      {
        label: 'First name',
        attribute: 'firstname',
      },
      {
        label: 'Last name',
        attribute: 'lastname',
        sorting: 'asc',
      },
      {
        label: 'Nickname',
        attribute: 'divisNickname',
      },
      {
        label: 'User name',
        attribute: 'username',
      },
      {
        label: 'Schools',
        attribute: 'schools',
        formatter: 'list',
        sortFormatter: 'list',
      },
      {
        label: 'Primary mail address',
        attribute: 'email',
      },
      {
        label: 'Activation state',
        attribute: 'disabled',
        formatter: 'boolean:Inactive:Active',
      },
      {
        label: 'Role',
        attribute: 'ucsschool_roles',
        formatter: 'list',
        sortFormatter: 'list',
      },
      {
        label: 'Source UID',
        attribute: 'source_uid',
      },
      {
        label: 'Record UID',
        attribute: 'record_uid',
      },
      {
        label: 'Creation date',
        attribute: 'create_timestamp',
        formatter: 'date:true',
      },
      {
        label: 'Date of last change',
        attribute: 'modify_timestamp',
        formatter: 'date:true',
      },
      {
        label: 'Account activation date',
        attribute: 'account_activation_date',
        formatter: 'date',
      },
      {
        label: 'Account expiry date',
        attribute: 'expiration_date',
      },
      {
        label: 'Scheduled deletion date',
        attribute: 'ucsschool_purge_timestamp',
      },
    ],
  };
  console.debug('mock config', _config);

  return _config;
};

const getError = async(response: Response): Promise<{detail: unknown} | null> => {
  if (!response.ok) {
    const text = await response.text();
    let detail = {
      detail: text,
    };
    try {
      const json = JSON.parse(text);
      if (json && typeof json === 'object' && 'detail' in json) {
        detail = json;
      }
    } catch (err) {
      // Either not a JSON response, or unparsable. We'll just ignore this.
    }
    return detail;
  }
  return null;
};

export const fetchListViewConfig = async(): Promise<ListViewConfig> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('fetchListViewConfig. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetchListViewConfigMocked());
      }, 1000);
    });
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/pageconf/listView');
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchListViewConfig.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
      unRecoverable: true,
    });
    return Promise.reject(answer);
  }
  return await answer.json();
};



const fetchAddViewConfigMocked = (): AddViewConfig => {
  const config: AddViewConfig = {
    pages: [
      {
        label: 'General',
        name: 'general',
        fieldsets: [
          {
            label: 'Basic settings',
            name: 'general__basicSettings',
            rows: [
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'firstname',
                    label: 'First name',
                    required: true,
                    access: 'write',
                  },
                },
                {
                  type: 'UInputText',
                  props: {
                    name: 'lastname',
                    label: 'Last name',
                    required: true,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'name',
                    label: 'Username',
                    required: true,
                    access: 'write',
                  },
                },
                {
                  type: 'UInputDate',
                  props: {
                    name: 'birthday',
                    label: 'Birthday',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'primaryMail',
                    label: 'Primary mail address',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputPassword',
                  props: {
                    name: 'password',
                    label: 'Password',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    label: 'Disabled',
                    name: 'disabled',
                    required: false,
                    access: 'write',
                  },
                },
              ],
            ],
          },
          {
            label: 'Account',
            name: 'general__account',
            rows: [
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'accountActivationDate',
                    label: 'Activation date',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'expirationDate',
                    label: 'Expiration date',
                    required: false,
                    access: 'write',
                  },
                },
              ],
            ],
          },
          {
            label: 'UCS@school',
            name: 'ucs@school',
            rows: [
              [
                {
                  type: 'UComboBox',
                  props: {
                    name: 'school',
                    label: 'Primary school',
                    required: true,
                    access: 'write',
                    options: [
                      {
                        label: 'School1',
                        value: 'school1',
                      },
                      {
                        label: 'School2',
                        value: 'school2',
                      },
                    ],
                  },
                },
              ],
              [
                {
                  type: 'UMultiSelect',
                  props: {
                    name: 'schoolClasses',
                    label: 'School classes',
                    required: true,
                    access: 'write',
                    columns: [
                      {
                        label: 'Select all',
                        attribute: 'label',
                      },
                    ],
                    rows: [
                      {
                        id: 'school1-1a',
                        attributes: {
                          label: {
                            value: '1A',
                            access: 'read',
                          },
                        },
                      },
                      {
                        id: 'school1-2a',
                        attributes: {
                          label: {
                            value: '2A',
                            access: 'read',
                          },
                        },
                      },
                      {
                        id: 'school1-3a',
                        attributes: {
                          label: {
                            value: '3A',
                            access: 'read',
                          },
                        },
                      },
                    ],
                  },
                },
              ],
              [
                {
                  type: 'UComboBox',
                  props: {
                    name: 'role',
                    label: 'Role',
                    required: true,
                    access: 'write',
                    options: [
                      {
                        value: 'student',
                        label: 'Lernende',
                      },
                      {
                        value: 'generic_user',
                        label: 'Nutzende',
                      },
                      {
                        value: 'teacher',
                        label: 'Lehrende',
                      },
                      {
                        value: 'school_admin',
                        label: 'Schuladmins',
                      },
                      {
                        value: 'technical_admin',
                        label: 'Fachliche Leitstelle',
                      },
                    ],
                  },
                },
              ],
            ],
          },
        ],
      },
    ],
  };
  console.debug('mock config', config);
  return config;
};
export const fetchAddViewConfig = async(): Promise<AddViewConfig> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('fetchAddViewConfig. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetchAddViewConfigMocked());
      }, 1000);
    });
  }
  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/pageconf/addView');
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchListViewConfig.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }

  return await answer.json();
};


const fetchUsersMocked = (): UserListResponseModel[] => {
  /* eslint-disable camelcase */
  const user: UserListResponseModel = {
    id: 'user1',
    allowedActions: ['edit', 'delete', 'setPassword', 'toggleWifi', 'toggleActive', 'exportCSV'],
    attributes: {
      lastname: {
        value: 'Mustermann',
        access: 'read',
      },
      firstname: {
        value: 'Max',
        access: 'read',
      },
      divisNickname: {
        value: 'Musti',
        access: 'read',
      },
      username: {
        value: 'user1',
        access: 'read',
      },
      name: {
        value: 'max.mustermann',
        access: 'read',
      },
      schools: {
        value: ['DEMOSCHOOL', 'school1', 'school2', 'school3'].sort(() => 0.5 - Math.random()),
        access: 'read',
      },
      school: {
        value: 'string string string string ',
        access: 'read',
      },
      email: {
        value: 'string',
        access: 'read',
      },
      disabled: {
        value: false,
        access: 'read',
      },
      ucsschool_roles: {
        value: ['role1'],
        access: 'read',
      },
      source_uid: {
        value: 'KELVIN',
        access: 'none',
      },
      record_uid: {
        value: 'mmrecord',
        access: 'none',
      },
      create_timestamp: {
        value: '2022-08-26T09:46:35.422Z',
        access: 'none',
      },
      modify_timestamp: {
        value: '2022-08-26T09:46:35.422Z',
        access: 'none',
      },
      account_activation_date: {
        value: '2022-08-26T09:46:35.422Z',
        access: 'none',
      },
      expiration_date: {
        value: '2022-08-26',
        access: 'none',
      },
      ucsschool_purge_timestamp: {
        value: '2022-08-26',
        access: 'none',
      },
    },
  };
  /* eslint-enable camelcase */

  const _users: UserListResponseModel[] = [user];

  for (let x = 2; x < 500; x++) {
    const y = JSON.parse(JSON.stringify(user));
    y.id = `user${x}`;
    y.attributes.account_activation_date.value = new Date(new Date(1900, 0, 1).getTime() + Math.random() * (new Date().getTime() - new Date(1900, 0, 1).getTime())).toDateString(); // eslint-disable-line no-mixed-operators
    y.attributes.schools.value = [...y.attributes.schools.value].sort(() => 0.5 - Math.random());
    y.attributes.username.value = y.id;
    if (x % 2 === 0) {
      y.allowedActions = ['edit'];
    }
    _users.push(y);
  }
  console.log('mock users: ', _users);
  return _users;
};

export const fetchUsers = async(
  school: string,
  group: string,
  role: string,
  sourceUID: string,
  state: string,
  quickSearch: string,
  networkAccess: string,
): Promise<UserListResponseModel[]> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('fetchUsers. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetchUsersMocked());
      }, 1000);
    });
  }

  const urlSearchParams = new URLSearchParams({school});
  if (group !== '') {
    urlSearchParams.append('group', group);
  }
  if (role !== '') {
    urlSearchParams.append('role', role);
  }
  if (sourceUID !== '') {
    urlSearchParams.append('sourceUID', sourceUID);
  }
  if (quickSearch !== '') {
    urlSearchParams.append('quickSearch', quickSearch);
  }
  if (state !== '') {
    urlSearchParams.append('disabled', (state === 'disabled').toString());
  }
  if (networkAccess) {
    urlSearchParams.append('networkAccess', (networkAccess === 'enabled').toString());
  }

  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/users?${urlSearchParams.toString()}`);
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchUsers.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }
  return await answer.json();
};

const fetchGroupsMocked = (): LabeledValue<string>[] => {
  const _groups = [
    {
      label: '1a',
      value: '1a',
    },
    {
      label: '2a',
      value: '2a',
    },
  ];
  console.log('mock groups: ', _groups);
  return _groups;
};

export const fetchGroups = async(school: string, groupKind: string, addGroupKindSuffixToLabel = false): Promise<LabeledValue<string>[]> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('fetchGroups. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetchGroupsMocked());
      }, 1000);
    });
  }
  const urlSearchParams = new URLSearchParams({school});
  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/${groupKind}?${urlSearchParams.toString()}`);
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchGroups.error.title', {
      context: groupKind,
    });
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }
  let groups: LabeledValue<string>[] = await answer.json();
  if (addGroupKindSuffixToLabel) {
    const groupKindSuffix = i18next.t(`ListView.searchForm.${groupKind}`);
    groups = groups.map(group => ({
      label: `${group.label} (${groupKindSuffix})`,
      value: group.value,
    }));
  }
  return groups;
};

const fetchUserMocked = (): UserDetailResponseModel => {
  const _user: UserDetailResponseModel = {
    url: '',
    allowedActions: ['save'],
    pages: [
      {
        label: 'General',
        name: 'general',
        fieldsets: [
          {
            label: 'Basic settings',
            name: 'general__basic',
            rows: [
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'firstname',
                    label: 'First name',
                    required: false,
                    access: 'write',
                  },
                },
                {
                  type: 'UInputText',
                  props: {
                    name: 'lastname',
                    label: 'Last name',
                    required: true,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'divisNameAffix',
                    label: 'Name affix',
                    required: false,
                  },
                },
                {
                  type: 'UInputText',
                  props: {
                    name: 'name',
                    label: 'Username',
                    required: true,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputPassword',
                  props: {
                    name: 'password',
                    label: 'Password',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'email',
                    label: 'Primary mail address',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    label: 'Disabled',
                    name: 'disabled',
                    required: false,
                    access: 'write',
                  },
                },
              ],
            ],
          },
          {
            label: 'Personal information',
            name: 'general__personalInformation',
            rows: [
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'divisNickname',
                    label: 'Nickname',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'displayName',
                    label: 'Display name',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'birthday',
                    label: 'Birthday',
                    required: false,
                    access: 'write',
                  },
                },
              ],
              [
                {
                  type: 'UMultiInput',
                  props: {
                    name: 'e-mail',
                    label: 'E-mails',
                    required: false,
                    subElements: [
                      {
                        type: 'UInputText',
                        props: {
                          name: 'e-mail__mail',
                          label: 'E-mails',
                        },
                      },
                    ],
                  },
                },
              ],
            ],
          },
          {
            label: 'UCS@school',
            name: 'general__ucs@school',
            rows: [
              [
                {
                  type: 'UComboBox',
                  props: {
                    name: 'school',
                    label: 'Primary school',
                    options: [
                      {
                        value: 'school1',
                        label: 'School 1',
                      }, {
                        value: 'school2',
                        label: 'School 2',
                      },
                    ],
                    required: true,
                  },
                },
              ],
              [
                {
                  type: 'UMultiSelect',
                  props: {
                    name: 'schools',
                    label: 'Schools',
                    columns: [
                      {
                        label: 'Select all',
                        attribute: 'label',
                      },
                    ],
                    rows: [
                      {
                        id: 'school1',
                        attributes: {
                          label: {
                            value: 'School 1',
                            access: 'read',
                          },
                        },
                      },
                      {
                        id: 'school2',
                        attributes: {
                          label: {
                            value: 'School 2',
                            access: 'read',
                          },
                        },
                      },
                    ],
                    required: true,
                  },
                },
              ],
              [
                {
                  type: 'UComboBox',
                  props: {
                    name: 'role',
                    label: 'Role',
                    required: true,
                    options: [
                      {
                        value: 'student',
                        label: 'Lernende',
                      },
                      {
                        value: 'generic_user',
                        label: 'Nutzende',
                      },
                      {
                        value: 'teacher',
                        label: 'Lehrende',
                      },
                      {
                        value: 'school_admin',
                        label: 'Schuladmins',
                      },
                      {
                        value: 'technical_admin',
                        label: 'Fachliche Leitstelle',
                      },
                    ],
                  },
                },
              ],
              [
                {
                  type: 'UComboBox',
                  props: {
                    name: 'sourceUID',
                    label: 'Source UID',
                    options: [
                      {
                        label: 'IDI',
                        value: 'IDI',
                      },
                      {
                        label: 'Manuell',
                        value: 'Manuell',
                      },
                    ],
                    required: false,
                  },
                },
                {
                  type: 'UInputText',
                  props: {
                    name: 'recordUID',
                    label: 'Record UID',
                    required: false,
                  },
                },
              ],
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'ucsschoolPurgeTimestamp',
                    label: 'Scheduled deletion date',
                    required: false,
                  },
                },
              ],
            ],
          },
        ],
      },
      {
        label: 'Account',
        name: 'account',
        fieldsets: [
          {
            label: 'General',
            name: 'account__general',
            rows: [
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'createTimestamp',
                    label: 'Creation date',
                    required: false,
                  },
                },
              ],
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'modifyTimestamp',
                    label: 'Date of last change',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'Deactivation',
            name: 'account__deactivation',
            rows: [
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'expirationDate',
                    label: 'Expiration date',
                    required: false,
                  },
                },
              ],
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    name: 'disabled',
                    label: 'Disabled',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'Locked login',
            name: 'account__lockedLogin',
            rows: [
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    name: 'pwdChangeNextLogin',
                    label: 'User has to change password on next login',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'Activation',
            name: 'account__activation',
            rows: [
              [
                {
                  type: 'UInputDate',
                  props: {
                    name: 'accountActivationDate',
                    label: 'Activation date',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'Password recovery',
            name: 'account__passwordRecovery',
            rows: [
              [
                {
                  type: 'UInputText',
                  props: {
                    name: 'PasswordRecoveryEmail',
                    label: 'Password recovery e-mail',
                    required: false,
                  },
                },
              ],
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    name: 'PasswordRecoveryEmailVerified',
                    label: 'Is the password recovery e-mail verified?',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'WLAN',
            name: 'account__wlan',
            rows: [
              [
                {
                  type: 'UInputCheckbox',
                  props: {
                    name: 'networkAccess',
                    label: 'WLAN (Radius) usage',
                    required: false,
                  },
                },
              ],
            ],
          },
          {
            label: 'SAML settings',
            name: 'account__samlSettings',
            rows: [
              [
                {
                  type: 'UMultiSelect',
                  props: {
                    name: 'serviceprovider',
                    label: 'SAML service provider',
                    required: false,
                    columns: [
                      {
                        label: 'Select all',
                        attribute: 'label',
                      },
                    ],
                    rows: [
                      {
                        id: 'sp1',
                        attributes: {
                          label: {
                            value: 'Service provider 1',
                            access: 'read',
                          },
                        },
                      },
                      {
                        id: 'sp2',
                        attributes: {
                          label: {
                            value: 'Service provider 2',
                            access: 'read',
                          },
                        },
                      },
                      {
                        id: 'sp3',
                        attributes: {
                          label: {
                            value: 'Service provider 3',
                            access: 'read',
                          },
                        },
                      },
                    ],
                  },
                },
              ],
            ],
          },
        ],
      },
    ],
    values: {
      'firstname': 'firstname',
      'lastname': 'lastname',
      'divisNameAffix': '',
      'name': 'username',
      'password': '',
      'email': '',
      'disabled': false,
      'divisNickname': 'nickname',
      'displayName': 'display name',
      'birthday': '2020-05-03',
      'e-mail': ['mustermann@max.de', 'max@muster.de'],
      'school': 'school1',
      'schools': ['school1'],
      'role': 'student',
      'sourceUID': '',
      'recordUID': '',
      'ucsschoolPurgeTimestamp': '',
      'createTimestamp': '',
      'modifyTimestamp': '',
      'expirationDate': '',
      'pwdChangeNextLogin': false,
      'accountActivationDate': '',
      'PasswordRecoveryEmail': '',
      'PasswordRecoveryEmailVerified': false,
      'networkAccess': false,
      'serviceprovider': [],
    },
  };
  console.debug('mock user', _user);
  return _user;
};

export const fetchUser = async(username: string): Promise<UserDetailResponseModel | null> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('fetchUser. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(fetchUserMocked());
      }, 1000);
    });
  }

  const answer = await fetchAuthenticated(`/ucsschool/bff-users/v1/users/${username}`);
  if (answer.status === 404) {
    return null;
  }
  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchUser.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }
  return await answer.json();
};


export type SaveUserError = {
  type: 'generic';
  message: string;
} | {
  type: 'fieldErrors';
  errors: {
    field: string;
    message: string;
  }[];
};

const getSaveUserError = async(response: Response, context: 'updateUser' | 'createUser'): Promise<SaveUserError | null> => {
  const errorJson = await getError(response);
  if (!errorJson) {
    return null;
  }
  if (response.status === 422) {
    // FIXME validate against json.schema
    if (Array.isArray(errorJson.detail)) {
      const detail: {loc: string[]; msg: string; type: string}[] = errorJson.detail;
      const error: SaveUserError = {
        type: 'fieldErrors',
        errors: [],
      };
      for (const iDetail of detail) {
        const field = iDetail.loc[iDetail.loc.length - 1];
        if (field) {
          error.errors.push({
            field,
            message: iDetail.msg,
          });
        }
      }
      return error;
    }
  }
  if (response.status === 404) {
    return {
      type: 'generic',
      message: i18next.t(`dataAccess.${context}.error.404`),
    };
  }
  return {
    type: 'generic',
    message: String(errorJson.detail),
  };
};
export const updateUser = async(url: string, values: FormValues): Promise<SaveUserError | null> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('updateUser. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    // TODO have the fail case in some way too
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(null);
        /*
        resolve({
          type: 'fieldErrors',
          errors: [
            {
              field: 'password',
              message: 'This should be overwritten in EditView',
            },
          ],
        });
        */
      }, 1000);
    });
  }

  const fixedProtocolUrl = url.replace(/https?:/, location.protocol);
  const answer = await fetchAuthenticated(`${fixedProtocolUrl}`, {
    method: 'PATCH',
    headers: {
      'Content-Type': 'application/json',
    },
    body: JSON.stringify(values),
  });
  return await getSaveUserError(answer, 'updateUser');
};

export const createUser = async(values: FormValues): Promise<SaveUserError | null> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('createUser. VITE_USE_REAL_BACKEND===false in DEV mode. Returning mock data');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve(null);
      }, 1000);
    });
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/users/', {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json',
    },
    body: JSON.stringify(values),
  });
  return await getSaveUserError(answer, 'createUser');

  // This commented-out block helped me identify issues that were occuring
  //  during e2e-testing only (cypress doesn't show the response body of
  //  failed requests)
  //
  // if (!answer.ok) {
  //   await new Promise((finishedReading): void => {
  //     let body = '';
  //     if (!answer.body) {
  //       console.log('No response body');
  //     } else {
  //       console.log('Reading response body');
  //       const reader = answer.body.getReader();
  //       const processBodyChunk = ({done, value}: ReadableStreamDefaultReadResult<Uint8Array>): Promise<unknown> => {
  //         body += new TextDecoder().decode(value);
  //         if (done) {
  //           finishedReading(body);
  //           console.log(body);
  //           return Promise.all([]);
  //         }
  //         return reader.read().then(processBodyChunk);
  //       };
  //       reader.read().then(processBodyChunk);
  //     }
  //   });
  // }
};

export const deleteUsers = async(usernames: string[]): Promise<{id: string; error: string}[]> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('deleteUsers. VITE_USE_REAL_BACKEND===false in DEV mode. Doing nothing');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve([
          {
            id: usernames[0] as string,
            error: 'some error',
          },
        ]);
      }, 1000);
    });
  }

  const responses = await Promise.all(usernames.map(username => fetchAuthenticated(`/ucsschool/bff-users/v1/users/${username}`, {
    method: 'DELETE',
  })));
  const errors = await Promise.all(responses.map(x => getError(x)));
  const fails: {id: string; error: string}[] = [];
  for (let x = 0; x < responses.length; x++) {
    const error = errors[x];
    if (error) {
      fails.push({
        id: usernames[x] as string,
        error: JSON.stringify(error.detail),
      });
    }
  }
  return fails;
};

interface CsvResponseModelSuccess {
  status: 'success';
  blob: Blob;
  filename: string;
}

export interface CsvResponseModelError {
  status: 'error';
  type: '422' | 'other';
  message: string;
}

type CsvResponseModel = CsvResponseModelError | CsvResponseModelSuccess;


export const exportUsers = async(usernames: string[]): Promise<CsvResponseModel> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('exportUsers. VITE_USE_REAL_BACKEND===false in DEV mode. Doing nothing');
    return new Promise(resolve => {
      setTimeout(() => {
        resolve({
          status: 'success',
          blob: new Blob(['"firstname","lastname"\n"Lars","Müller"'], {type: 'text/csv'}),
          filename: 'export-user.csv',
        });
      }, 1000);
    });
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/csv', {
    method: 'POST',
    headers: {
      'accept': 'text/csv',
      'Content-Type': 'application/json',
    },
    body: JSON.stringify({users: usernames}),
  });

  const error = await getError(answer);
  if (error) {
    return {
      status: 'error',
      type: answer.status === 422 ? '422' : 'other',
      message: JSON.stringify(error.detail),
    };
  }

  const content = await answer.blob();
  const filename = await answer.headers.get('Content-Disposition');
  const matches = filename && /filename\s*=\s*(?<filename>[\S\s]+)/g.exec(filename);
  const extractedName = matches && matches.length > 0 && matches[1];

  return {
    status: 'success',
    blob: content,
    filename: extractedName ? extractedName : 'user-export.csv',
  };
};


interface BackendSetPasswordsResponse {
  failedUsers: {id: string; error: string}[];
  file: {
    filename: string;
    data: string;
  };
}
interface SetPasswordResponseModelSuccess {
  status: 'success';
  response: BackendSetPasswordsResponse;
}
export interface SetPasswordResponseModelError {
  status: 'error';
}
type SetPasswordResponseModel = SetPasswordResponseModelError | SetPasswordResponseModelSuccess;
export const setPasswords = async(password: string, usernames: string[]): Promise<SetPasswordResponseModel> => {
  if (import.meta.env.VITE_USE_REAL_BACKEND !== 'true') {
    console.debug('setPasswords. VITE_USE_REAL_BACKEND===false in DEV mode. Doing nothing');
    return new Promise(resolve => {
      setTimeout(() => {
        /*
        resolve({
          status: 'error',
        });
        */
        resolve({
          status: 'success',
          response: {
            failedUsers: [
              /*
              {id: 'user1', error: 'some error'},
              {id: 'user2222', error: 'some other error'},
              */
            ],
            file: {
              filename: 'set-password.xls',
              data: 'SGVyZSBjb3VsZCBiZSB4bHN4IGNvbnRlbnQsIGJ1dCB0aGVyZSBpc24ndC4gVGhpcyBpcyBqdXN0IGEgbW9jay4=',
            },
          },
        });

      }, 1000);
    });
  }

  const answer = await fetchAuthenticated('/ucsschool/bff-users/v1/set-passwords', {
    method: 'POST',
    headers: {
      'accept': 'application/json',
      'Content-Type': 'application/json',
    },
    body: JSON.stringify({
      users: usernames,
      password,
    }),
  });

  const error = await getError(answer);
  if (error) {
    const errorStore = useErrorStore();
    const title = i18next.t('dataAccess.fetchListViewConfig.error.title');
    errorStore.push({
      title,
      message: JSON.stringify(error.detail),
    });
    return Promise.reject(answer);
  }

  const response: BackendSetPasswordsResponse = await answer.json();
  return {
    status: 'success',
    response,
  };
};
